/*
 * Copyright (C) 2006-2012 Frederic Saunier, www.tekool.net
 * Objs may be freely distributed under the MIT license.
 * For all details and documentation: https://github.com/tekool/objs/
 */
var	$InvalidClasspath/*String*/ = "invalid classpath: ";
var	$NonExistentClass/*String*/ = "non-existent class: ";
var	$NonExistentSuperClass/*String*/ = "non-existent superclass: ";
var	$InvalidProtobject/*String*/ = "invalid protobject for: ";

var classpath/*String*/ = "MyClass";
var superclasspath/*String*/ = "MySuperClass";
var subclasspath/*String*/ = "MySubClass";
var sub2classpath/*String*/ = "MySub2Class";

/**
 * @classDescription
 * Test the Objs library methods.
 */
var ObjsTest = new YUITest.TestCase
(
	{
	    /**
	     * The name of the test case - if not provided, one is automatically
	     * generated by the YUITest framework.
	     * 
	     * @type {String}
	     * @private
	     */
	    name: "ObjsTest library tests",  

	    /**
	     * Sets up data that is needed by each test.
	     */
	    setUp: function(){},

	    /**
	     * Cleans up everything that was created by setUp().
	     */
	    tearDown: function()
	    {
	    	Objs( classpath, null );
	    	Objs( superclasspath, null );
	    	Objs( subclasspath, null );
	    	Objs( sub2classpath, null );
	    },
		
		/**
		 * Tests Objs(classpath) parameter.
		 */
		testRetrieveInvalidClasspathParameter: function()
		{
			YUITest.Assert.throwsError
			(
				$InvalidClasspath + undefined,
				function(){Objs()},
				"An error should have been thrown."
			);
			
			YUITest.Assert.throwsError
			(
				$InvalidClasspath + "null",
				function(){Objs(null)},
				"An error should have been thrown."
			);

			YUITest.Assert.throwsError
			(
				$InvalidClasspath + "undefined",
				function(){Objs(undefined)},
				"An error should have been thrown."
			);

			//Not a string
			YUITest.Assert.throwsError
			(
				$InvalidClasspath + "[object Object]",
				function(){Objs({})},
				"An error should have been thrown."
			);
		},

		/**
		 * Tests Objs(classpath) class constructor creation.
		 */
		testRetrieveNonExistent: function()
		{
			YUITest.Assert.throwsError
			(
				$NonExistentClass + classpath,
				function(){Objs( classpath )},
				"An error should have been thrown."
			);
		},

		/**
		 * Tests Objs method invalid classpath argument errors
		 */
		testCreateInvalidClasspathParameter: function()
		{
			YUITest.Assert.throwsError
			(
				$InvalidClasspath + "null",
				function(){Objs(null,{})},
				"An error should have been thrown."
			);
			
			YUITest.Assert.throwsError
			(
				$InvalidClasspath + "undefined",
				function(){Objs(undefined,{})},
				"An error should have been thrown."
			);
			
			//Not a string
			YUITest.Assert.throwsError
			(
				$InvalidClasspath + "[object Object]",
				function(){Objs({},{})},
				"An error should have been thrown."
			);
		},

		/**
		 * Tests Objs(classpath) class constructor creation with an empty
		 * protobject.
		 */
		testCreateWithObject: function()
		{
			var constructor/*Function*/ = Objs( classpath, {} );

			YUITest.Assert.isFunction
			(
				constructor,
				"Expected constructor to be an instance of Function"
			);
		},

		/**
		 * Tests Objs(classpath) class constructor creation with an empty
		 * function as protobject.
		 */
		testCreateWithFunction: function()
		{
			var constructor/*Function*/ = Objs( classpath, new function(){} );

			YUITest.Assert.isFunction
			(
				constructor,
				"Expected constructor to be an instance of Function"
			);
		},

		/**
		 * Tests Objs(classpath) class removal.
		 */
		testRemove: function()
		{
			var constructor/*Function*/ = Objs( classpath, {} );

			YUITest.Assert.isFunction
			(
				constructor,
				"Expected constructor to be an instance of Function"
			);
			
			var removedConstructor = Objs(classpath,null);

			YUITest.Assert.areEqual
			(
				constructor,
				removedConstructor,
				"Expected added constructor to be equals to removedConstructor"
			);
			
			YUITest.Assert.throwsError
			(
				$NonExistentClass + classpath,
				function(){Objs(classpath)},
				"An error should have been thrown."
			);
		},

		/**
		 * Tests that each class owns its classPath property.
		 */
		testClassPathExists: function()
		{
			YUITest.Assert.throwsError
			(
				$NonExistentClass + classpath,
				function(){Objs( classpath )},
				"An error should have been thrown."
			);
		},

		/**
		 * Tests Objs method invalid superclass argument errors.
		 */
		testInheritingWithInvalidParameters: function()
		{
			YUITest.Assert.throwsError
			(
				$NonExistentSuperClass,
				function(){Objs(classpath,"", {})},
				"An error should have been thrown."
			);

            YUITest.Assert.throwsError
            (
                $NonExistentSuperClass + superclasspath,
                function(){Objs( classpath, superclasspath, {} )},
                "An error should have been thrown."
            );

            Objs(superclasspath,{});

            YUITest.Assert.throwsError
            (
                $InvalidProtobject + classpath,
                function(){Objs( classpath, superclasspath )},
                "An error should have been thrown."
            );

            YUITest.Assert.throwsError
            (
                $InvalidProtobject + classpath,
                function(){Objs( classpath, superclasspath, null )},
                "An error should have been thrown."
            );
        },

		/**
		 * Tests Objs() using Objs() method.
		 */
		testCreateAndGet: function()
		{
			var constructor/*Function*/ = Objs( classpath, {} );
			
			YUITest.Assert.areEqual
			(
				constructor,
				Objs(classpath),
				"Expected Objs( classpath ) to return the registered constructor"
			);
		},

		/**
		 * Tests Objs(classpath) called twice with the same classpath.
		 */
		testCreateTwice: function()
		{
			var constructor1/*Function*/ = Objs( classpath, {} );
			var constructor2/*Function*/ = Objs( classpath, {} );
			
			YUITest.Assert.areNotEqual
			(
				constructor1,
				constructor2,
				"Expected constructor1 and constructor2 to be the same"
			);
			
			YUITest.Assert.areEqual
			(
				constructor2,
				Objs( classpath ),
				"Expected constructor1 and constructor2 to be the same"
			);
		},
		
		/**
		 * Tests Objs() method <code>protobject</code> used as 2nd argument.
		 */
		testProtobject: function()
		{
			var MyClass/*Function*/ = Objs
			(
				classpath,
				{
					myProp: 41,
					myFunc: function()
					{
						this.myProp++;
					}
				}
			);
			
			var myClass/*MyClass*/ = new MyClass();
			myClass.myFunc();

			YUITest.Assert.areEqual
			(
				42,
				myClass.myProp,
				"Expected myClass.myProp to be 42"
			);
		},
		
		/**
		 * Tests Objs() method with <code>SuperClass</code> used as 2nd
		 * argument.
		 */
		testSuperClassAsFunction: function()
		{
			var SuperClass/*Function*/ = Objs( superclasspath, {} );
			var SubClass/*Function*/ = Objs( subclasspath, SuperClass, {} );

			YUITest.Assert.isInstanceOf
			(
				SuperClass,
				new SubClass(),
				"Expected SubClass to be an instance of SuperClass"
			);
			
			var sub2class/*Function*/ = Objs( sub2classpath, SubClass, {} );

			YUITest.Assert.isInstanceOf
			(
				SuperClass,
				new sub2class(),
				"Expected sub2class to be an instance of SuperClass"
			);
		},
		
		/**
		 * Tests Objs() method with <code>SuperClass</code> used as String
		 * as 2nd argument.
		 */
		testSuperClassAsString: function()
		{
			var SuperClass/*Function*/ = Objs( superclasspath, {} );
			var SubClass/*Function*/ = Objs( subclasspath, superclasspath, {} );

			YUITest.Assert.isInstanceOf
			(
				SuperClass,
				new SubClass(),
				"Expected SubClass to be an instance of SuperClass"
			);
			
			var sub2class/*Function*/ = Objs( sub2classpath, subclasspath, {} );

			YUITest.Assert.isInstanceOf
			(
				SuperClass,
				new sub2class(),
				"Expected sub2class to be an instance of SuperClass"
			);
		},
			
		/**
		 * Tests Objs() method with <code>SuperClass</code> used as 2nd
		 * argument and <code>protobject</code> used as 3nd argument.
		 */
		testSuperClassPlusProtobject: function()
		{
			var SuperClass/*Function*/ = Objs
			(
				subclasspath,
				{ 
					myProp: 41
				}
			);
			
			var SubClass/*Function*/ = Objs
			(
				subclasspath,
				SuperClass,
				{
					myFunc: function()
					{
						this.myProp++;
					}
				}
			);

			YUITest.Assert.isInstanceOf
			(
				SuperClass,
				new SubClass(),
				"Expected SubClass to be an instance of SuperClass"
			);

			var subClass/*SubClass*/ = new SubClass();
			subClass.myFunc();

			YUITest.Assert.areEqual
			(
				42,
				subClass.myProp,
				"Expected subClass.myProp to be 42"
			);
		},

		/**
		 * Tests if the initialize method of each class is called and respect
		 * the recursion order.
		 */
		testInitialize: function()
		{
			var proof/*String*/ = "";
			var SuperClass/*Function*/ = Objs(superclasspath, {});
			SuperClass.prototype.initialize = function()
			{
				proof += "1";
			};

			var SubClass/*Function*/ = Objs( subclasspath, SuperClass, {} );
			SubClass.prototype.initialize = function()
			{
				SubClass.$super.initialize.call(this);
				proof += "2";
			};

			var Sub2class/*Function*/ = Objs( sub2classpath, SubClass, {} );
			Sub2class.prototype.initialize = function()
			{
				Sub2class.$super.initialize.call(this);
				proof += "3";
			};
			
			new Sub2class();

			YUITest.Assert.areEqual
			(
				"123",
				proof,
				"Expected intialize to be called recursively in the right order on each method"
			);
		},

		/**
		 * Tests if the initialize method of each class is called and respect
		 * the recursion order.
		 */
		testInitializeParameters: function()
		{
			var proof/*String*/ = "";
			var SuperClass/*Function*/ = Objs(superclasspath, {});
			SuperClass.prototype.initialize = function(arg)
			{
				proof += arg;
			};

			var SubClass/*Function*/ = Objs( subclasspath, SuperClass, {} );
			SubClass.prototype.initialize = function(arg)
			{
				SubClass.$super.initialize.call( this, "1" );
				proof += arg;
			};

			var Sub2class/*Function*/ = Objs( sub2classpath, SubClass, {} );
			Sub2class.prototype.initialize = function(arg)
			{
				Sub2class.$super.initialize.call( this, "2" );
				proof += arg;
			};
			
			new Sub2class("3");

			YUITest.Assert.areEqual
			(
				"123",
				proof,
				"Expected intialize to be called with arguments recursively in the right order on each method"
			);
		},

		/**
		 * Tests if the initialize method respect recursion call order even if
		 * the inheritance chain is broken by a subclass not declaring its
		 * initialize method.
		 */
		testInitializeNotDeclaredInChain: function()
		{
			var proof/*String*/ = "";
			var SuperClass/*Function*/ = Objs(superclasspath, {});
			SuperClass.prototype.initialize = function()
			{
				proof += "1";
			};

			var SubClass/*Function*/ = Objs( subclasspath, SuperClass, {} );
			var Sub2class/*Function*/ = Objs( sub2classpath, SubClass, {} );
			Sub2class.prototype.initialize = function()
			{
				Sub2class.$super.initialize.call(this);
				proof += "2";
			};

			new Sub2class();

			YUITest.Assert.areEqual
			(
				"12",
				proof,
				"Expected intialize to be called with arguments recursively in the right order on each method"
			);
		},
		
		/**
		 * Tests if the initialize method respect recursion call order even if
		 * the inheritance chain is broken by the first class in the prototype
		 * chain not declaring its initialize method.
		 */
		testInitializeNotDeclared: function()
		{
			var proof/*String*/ = "";
			var SuperClass/*Function*/ = Objs(superclasspath, {});
			SuperClass.prototype.initialize = function()
			{
				proof += "1";
			};

			var SubClass/*Function*/ = Objs( subclasspath, SuperClass, {} );
			SubClass.prototype.initialize = function()
			{
				SubClass.$super.initialize.call(this);
				proof += "2";
			};
			
			var Sub2class/*Function*/ = Objs( sub2classpath, SubClass, {} );
			
			
			//TODO Need to be explored ... there's something still not well managed that happens here
			new Sub2class();

			YUITest.Assert.areEqual
			(
				"12",
				proof,
				"Expected intialize to be called with arguments recursively in the right order on each method"
			);
		},
		
		/**
		 * Tests if the initialize method respect recursion call order even if
		 * the inheritance chain is broken by a subclass not declaring its
		 * initialize method.
		 */
		testInitializeThis: function()
		{
			var SuperClass/*Function*/ = Objs(superclasspath, {});
			SuperClass.prototype.initialize = function()
			{
				this.proof += "1";
			};

			var SubClass/*Function*/ = Objs( subclasspath, SuperClass, {} );
			SubClass.prototype.initialize = function()
			{
				SubClass.$super.initialize.call(this);
				this.proof += "2";
			};
			
			var Sub2class/*Function*/ = Objs( sub2classpath, SubClass, {} );
			Sub2class.prototype.proof = "0";
			
			//TODO Need to be explored ... there's something still not well managed that happens here
			var testClass/*Sub2class*/ = new Sub2class();

			YUITest.Assert.areEqual
			(
				"012",
				testClass.proof,
				"Expected intialize to be called with arguments recursively in the right order on each method"
			);
		},
	
		/**
		 * Tests if we can call the superclass methods from a subclass.
		 */
		testSuperMethods: function()
		{
			var SuperClass/*Function*/ = Objs
			(
				superclasspath,
				{
					proof : "",

					method1 : function()
					{
						this.proof += "SuperClassMethod1";
					},
				
					method2 : function()
					{
						this.proof += "SuperClassMethod2";
					}
				}
			);

			var SubClass/*Function*/ = Objs
			(
				subclasspath,
				SuperClass,
				{
					method1 : function()
					{
						SubClass.$super.method1.call(this);
						
						this.proof += "SubClassMethod1";
					},
				
					method2 : function()
					{
						SubClass.$super.method2.call(this);
						
						this.proof += "SubClassMethod2";
					}
				}
			);
			
			//TODO Need to be explored ... there's something still not well managed that happens here
			var testClass/*SubClass*/ = new SubClass();
			testClass.method1();
			testClass.method2();

			YUITest.Assert.areEqual
			(
				"SuperClassMethod1SubClassMethod1SuperClassMethod2SubClassMethod2",
				testClass.proof,
				"Expected methods to be called recursively in the right order"
			);
		},
		
		/**
		 * Tests if a class can override non enumerable methods from Object.
		 */
		testNonEnumerable: function()
		{
			var MyClass/*Function*/ = Objs
			(
				classpath,
				{
					toString : function()
					{
						return "toString() - okay";
					},
					
					toLocaleString : function()
					{
						return "toLocaleString() - okay";
					},
					
					valueOf : function()
					{
						return "valueOf() - okay";
					}
				}
			);
			
			var testClass/*MyClass*/ = new MyClass();

			YUITest.Assert.areEqual
			(
				"toString() - okay",
				testClass.toString(),
				'Expected "toString - okay"'
			);

			YUITest.Assert.areEqual
			(
				"toLocaleString() - okay",
				testClass.toLocaleString(),
				'Expected "toLocaleString - okay"'
			);

			YUITest.Assert.areEqual
			(
				"valueOf() - okay",
				testClass.valueOf(),
				'Expected "valueOf - okay"'
			);
		}
  	}
);